/*
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights reserved.
 *
 * Author: Pops, TCSDeveloper
 */

using System;
using System.Collections;

namespace TopCoder.Rules.DependencyMap.View
{

    /// <summary>
    /// This is the main class of the application. This class will be used by the application to:
    /// 1) Manipulate the layout manager that will be used to layout the dependency diagrams. Only a single
    /// layout manager can be active at any given time.
    /// 2) Manipulate the renderer(s) used to render the dependency diagrams. More than one renderer can be used.
    /// 3) Provide a dependency and have it layed out into a diagram.
    /// </summary>
    public class DependencyRenderer
    {

        /// <summary>
        /// The shape factory used by this renderer to produce the shapes used in the rendering process. This
        /// variable is set in the constructor, is immutable (the reference only) once set and will be used in
        /// the render method. This variable has been marked readonly to enforce it's immutablility (again -
        /// reference only).
        /// </summary>
        private readonly ShapeFactory shapeFactory;

        /// <summary>
        /// The variable represents the renderes that will be used to render the diagram. Each element in
        /// the list will be of DiagramRenderer type. This value is set during class construction time,
        /// is immutable (the reference not the contents) and will be used in the render method. The list
        /// content can be accessed via the addDiagramRenderer(), removeDiagramRenderer() and DiagramRenderers
        /// property. This variable has been marked readonly to ensure it's immutability (again reference only).
        /// </summary>
        private readonly IList renderers = new ArrayList();

        /// <summary>
        /// The variable represents the layout manager to use when laying the dependency diagram out. This variable
        /// will initially be set at object construction time, is mutatable by the setter method and will be
        /// accessed in both the getter method and the render method.
        /// </summary>
        private ILayoutManager layoutManager = new FixedGridLayoutManager();

        /// <summary>
        /// This is the property for the layoutManager variable. This is a simple setter/getter for the
        /// layoutManager variable.
        /// </summary>
        /// <exception cref="ArgumentNullException">if value is null</exception>
        public ILayoutManager LayoutManager
        {
            get
            {
                return this.layoutManager;
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("layoutManager");
                }
                this.layoutManager = value;
            }
        }

        /// <summary>
        /// This is the property for the renderers variable. This property defines ONLY the getter (no setter).
        /// This property getter will return a ReadOnly version of the renderers.
        /// </summary>
        public IList DiagramRenderers
        {
            get
            {
                return ArrayList.ReadOnly(this.renderers);
            }
        }

        /// <summary>
        /// Constructs the DependencyRenderer with a default ShapeFactory. This method will simply call
        /// 'this(new ShapeFactory())'.
        /// </summary>
        public DependencyRenderer() : this(new ShapeFactory())
        {
        }

        /// <summary>
        /// Constructs the DependencyRenderer from the specified shapeFactory. This method will simply save
        /// the shapeFactory reference to the like named variable.
        /// </summary>
        /// <param name="shapeFactory">A non-null shapeFactory to use.</param>
        /// <exception cref="ArgumentNullException">if shapeFactory is null</exception>
        public DependencyRenderer(ShapeFactory shapeFactory)
        {
            if (shapeFactory == null)
            {
                throw new ArgumentNullException("shapeFactory");
            }
            this.shapeFactory = shapeFactory;
        }

        /// <summary>
        /// Adds a new renderer to the list of renderers that will be used to render a dependency diagram.
        /// This method simply adds the renderer to the renderers list.
        /// </summary>
        /// <param name="renderer">the renderer to add to the list of renderers</param>
        /// <exception cref="ArgumentNullException">if renderer is null</exception>
        public void AddDiagramRenderer(IDiagramRenderer renderer)
        {
            if (renderer == null)
            {
                throw new ArgumentNullException("renderer");
            }
            this.renderers.Add(renderer);
        }

        /// <summary>
        /// This method will remove the specified renderer from the list of renderers.
        /// </summary>
        /// <param name="renderer">The renderer to remove</param>
        /// <exception cref="ArgumentNullException">if renderer is null</exception>
        public void RemoveDiagramRenderer(IDiagramRenderer renderer)
        {
            if (renderer == null)
            {
                throw new ArgumentNullException("renderer");
            }
            this.renderers.Remove(renderer);
        }

        /// <summary>
        /// This method is the method that will be used by the application to render the dependency diagram.
        /// The layout manager will be used to layout the diagram against each renderer (note: because of
        /// differences in the font metrics between renderers, the layout must be repeated for each renderer
        /// to account for those differences). This method will:
        /// 1) Iterate over the renderers and for each renderer ...
        /// 2) Call the layoutManager.Layout(dependency, shapeFactory, renderer) to layout the diagram.
        /// If an RendererException is thrown, simply let it pass through to the caller.
        /// 3) Repeat until all renderers are done
        /// </summary>
        /// <param name="dependency">The dependency that will be rendered</param>
        /// <exception cref="ArgumentNullException">if dependency is null</exception>
        /// <exception cref="RendererException">if an exception occurs during rendering</exception>
        public void Render(Dependency dependency)
        {
            if (dependency == null)
            {
                throw new ArgumentNullException("dependency");
            }

            foreach (IDiagramRenderer renderer in this.renderers)
            {
                layoutManager.Layout(dependency, shapeFactory, renderer);
            }
        }
        
        /// <summary>
        /// This method will render all unique nodes within the dependency map - including the top level
        /// node.  This method simply calls RenderAll(dependency, true).  The diagram renderer used should
        /// be able to support multiple diagrams being rendered into one instance of itself.
        /// </summary>
        /// <param name="dependency">The dependency that will be rendered.</param>
        /// <exception cref="ArgumentNullException">If dependency is null.</exception>
        /// <exception cref="RendererException">If an exception oocurs in the renderer.</exception>
        public void RenderAll(Dependency dependency)
        {
        	// Dependency will be checked in RenderAll(Dependency, bool)
        	
        	this.RenderAll(dependency, true);
        }
        
        /// <summary>
        /// This method will render all unique nodes within the dependency map
		/// - including the top level node if renderTopLevel is true.  The diagram renderer used should be able
		/// to support multiple diagrams being rendered into one instance of itself.
        /// </summary>
        /// <param name="dependency"></param>
        /// <param name="renderTopLevel"></param>
        /// <exception cref="ArgumentNullException">If dependency is null.</exception>
        /// <exception cref="RendererException">If an exception oocurs in the renderer.</exception>
        public void RenderAll(Dependency dependency, bool renderTopLevel)
        {
        	if (dependency == null)
        	{
        		throw new ArgumentNullException("dependency");
        	}

        	Queue queue = new Queue();

        	if (renderTopLevel)
        	{
        		queue.Enqueue(dependency);
        	} else
        	{
	        	foreach (Dependency child in dependency.Dependants)
    	    	{
	        		queue.Enqueue(child);
        		}	
        	}

        	Hashtable seen = new Hashtable();
        	
        	// Do a breadth first search on the children, rendering only the unseen ones
        	while (queue.Count > 0)
        	{
        		Dependency current = (Dependency)queue.Dequeue();
        		
        		if (seen.Contains(current.Title))
        		{
        			// Onto next dependency if we have seen this already
        			continue;
        		}

        		seen[current.Title] = true;

        		this.Render(current);

	        	foreach (Dependency child in current.Dependants)
    	    	{
	        		queue.Enqueue(child);
        		}	
        	}
        }

    }

}
