/*
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights reserved.
 *
 * Author: Pops, TCSDeveloper
 */

using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace TopCoder.Rules.DependencyMap.View
{

    /// <summary>
    /// The shape factory will be used by the component to generate shapes that will be used when laying out
    /// the diagram. A shape is defined by it's GraphicsPath. This class allows the application to override
    /// the 'builtin' shapes and to specify new shapes to be used (see the Dependency class). Shapes put into this
    /// factory will be Cloned() and potentially resized/rotated by the layout manager. This class will ensure that
    /// the 'common' shapes (as defined by the constants) exist (ie the component can rely on this fact).
    /// </summary>
    public class ShapeFactory
    {

        /// <summary>
        /// This constant field defines the 'common' shape that will be used to represent relationship arrows
        /// (a relationship between dependencies).
        /// </summary>
        public const int RelationshipArrow = 1;

        /// <summary>
        /// This constant field defines the 'common' shape that will be used to represent a dependency element
        /// that has not been 'seen' yet (see algo documentation for more details).
        /// </summary>
        public const int UnseenElement = 2;

        /// <summary>
        /// This constant field defines the 'common' shape that will be used to represent a dependency element
        /// that has been 'seen' already (see algo documentation for more details).
        /// </summary>
        public const int SeenElement = 3;

        /// <summary>
        /// This dictionary is used to store the type (key) to it's corresponding shape (value). The map is
        /// initialized at object construction time, is immutable (the reference, not the contents) and will be
        /// manipulated by every method! This variable has been marked readonly to enforce it's immutablility
        /// (again - reference only).
        /// </summary>
        private readonly IDictionary shapes = new Hashtable();

        /// <summary>
        /// This constructor defines the ShapeFactory with the common types. This method will create the following
        /// shapes and store them into the shapes:
        /// 1) RELATIONSHIP_ARROW - will create a GraphicsPath that describes the arrow that is a close proximity
        /// to those found in the requirements. The path should fit in a 20 (width), 40 (height) square. The arrow
        /// should be pointing down.
        /// 2) UNSEEN_ELEMENT - will create a square of size 40 via the GraphicsPath.
        /// 3) SEEN_ELEMENT - will create a GraphicsPath that describes a rounded square of size 40.
        /// Please note that the numbers above are pure estimates and the programmer can change the defaults to more
        /// appropriate values.
        /// </summary>
        public ShapeFactory()
        {
            // Create a GraphicsPath for Relationship Arrow.
            GraphicsPath relationshipArrow = new GraphicsPath();
            relationshipArrow.AddLine(14, 0, 14, 30);
            relationshipArrow.AddLine(14, 30, 19, 30);
            relationshipArrow.AddLine(19, 30, 10, 40);
            relationshipArrow.AddLine(10, 40, 0, 30);
            relationshipArrow.AddLine(0, 30, 5, 30);
            relationshipArrow.AddLine(5, 30, 5, 0);
            relationshipArrow.CloseFigure();
            shapes[ShapeFactory.RelationshipArrow] = relationshipArrow;

            // Create a GraphicsPath for Unseen Element.
            GraphicsPath unseenElement = new GraphicsPath();
            unseenElement.AddRectangle(new Rectangle(0, 0, 40, 40));
            shapes[ShapeFactory.UnseenElement] = unseenElement;

            // Create a GraphicsPath for Seen Element.
            GraphicsPath seenElement = new GraphicsPath();
            seenElement.AddCurve(new Point[] {new Point(34, 0), new Point(38, 1), new Point(39, 5)});
            seenElement.AddLine(39, 5, 39, 34);
            seenElement.AddCurve(new Point[] {new Point(39, 34), new Point(38, 38), new Point(34, 39)});
            seenElement.AddLine(34, 39, 5, 39);
            seenElement.AddCurve(new Point[] {new Point(5, 39), new Point(1, 38), new Point(0, 34)});
            seenElement.AddLine(0, 34, 0, 5);
            seenElement.AddCurve(new Point[] {new Point(0, 5), new Point(1, 1), new Point(5, 0)});
            seenElement.CloseFigure();
            shapes[ShapeFactory.SeenElement] = seenElement;
        }

        /// <summary>
        /// This method is the indexer for the shapes variable and defines both setter and getter actions. The
        /// getter should return null if the type was not found and a Clone() of the GraphicsPath for the type
        /// if found. The setter should create the type/graphicspath relation if it doesn't currently exist or
        /// replaces the existing graphicspath if the type is found.
        /// </summary>
        /// <param name="type">An integer representing the unique type</param>
        /// <returns>The shape that was identified by the type or null if it was a new shape</returns>
        /// <exception cref="ArgumentNullException">if the value is null</exception>
        /// <exception cref="ArgumentOutOfRangeException">if type is less than or equal to 0</exception>
        public GraphicsPath this[int type]
        {
            get
            {
                if (type <= 0)
                {
                    throw new ArgumentOutOfRangeException("type should be positive.");
                }
                GraphicsPath shape = this.shapes[type] as GraphicsPath;
                if (shape == null)
                {
                    return null;
                }
                return shape.Clone() as GraphicsPath;
            }
            set
            {
                if (type <= 0)
                {
                    throw new ArgumentOutOfRangeException("type should be positive.");
                }
                if (value == null)
                {
                    throw new ArgumentNullException("shape");
                }
                this.shapes[type] = value;
            }
        }

        /// <summary>
        /// This method allows the application to add a new shape into the shape factory.
        /// </summary>
        /// <param name="type">An integer representing the unique type</param>
        /// <param name="shape">an shape The shape to relate to the type</param>
        /// <exception cref="ArgumentNullException">if the shape is null</exception>
        /// <exception cref="ArgumentException">if the type already exists</exception>
        /// <exception cref="ArgumentOutOfRangeException">if the type is less than or equal to zero</exception>
        public void AddShape(int type, GraphicsPath shape)
        {
            if (this.shapes.Contains(type))
            {
                throw new ArgumentException("type already exists.");
            }
            this[type] = shape;
        }

        /// <summary>
        /// This method will remove the specified shape from the shape factory. The specified type cannot be on
        /// of the common types (as defined by the constants).
        /// </summary>
        /// <param name="type">The type that identifies the shape to remove</param>
        /// <exception cref="ArgumentException">if the type is one of the predefined types</exception>
        /// <exception cref="ArgumentOutOfRangeException">if the type is less than or equal to zero</exception>
        public void RemoveShape(int type)
        {
            if (type <= 0)
            {
                throw new ArgumentOutOfRangeException("type should be positive.");
            }
            if (type == ShapeFactory.RelationshipArrow || type == ShapeFactory.UnseenElement ||
                type == ShapeFactory.SeenElement)
            {
                throw new ArgumentException("Pre-defined type should not be removed.");
            }
            this.shapes.Remove(type);
        }

    }

}
