/**
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights Reserved.
 */
package com.topcoder.rules.dependencymap.view;

import java.awt.geom.GeneralPath;
import java.util.Map;
import java.util.HashMap;

/**
 * The shape factory will be used by the component to generate shapes that will be used when laying out the diagram.
 * This class allows the application to override the 'builtin' shapes and to specify new shapes to be used (see the
 * Dependency class). Shapes put into this factory will be 'cloned' (the shape must implement Cloneable) and
 * potentially resized/rotated by the layout manager. This class will ensure that the 'common' shapes (as defined
 * by the constants) exist (ie the component can rely on this fact).
 *
 * @author Pops
 * @author TCSDeveloper
 */
public class ShapeFactory {

    /**
     * This constant field defines the 'common' shape that will be used to represent relationship arrows
     * (a relationship between dependencies).
     */
    public static final int RELATIONSHIP_ARROW = 1;

    /**
     * This constant field defines the 'common' shape that will be used to represent a dependency element that
     * has not been 'seen' yet (see algo documentation for more details).
     */
    public static final int UNSEEN_ELEMENT = 2;

    /**
     * This constant field defines the 'common' shape that will be used to represent a dependency element that
     * has been 'seen' already (see algo documentation for more details).
     */
    public static final int SEEN_ELEMENT = 3;

    /**
     * This map is used to store the type (Integer key) to it's corresponding shape (value). The map is initialized
     * at object construction time, is immutable (the reference, not the contents) and will be manipulated by every
     * method! This variable has been marked final to enforce it's immutablility (again - reference only).
     */
    private final Map shapes = new HashMap();

    /**
     * This constructor defines the ShapeFactory with the common types. This method will create the following shapes
     * and store them into the shapes:
     *
     * 1) RELATIONSHIP_ARROW - will create a GeneralPath that describes the arrow that is a close proximity to those
     * found in the requirements. The path should fit in a 20 (width), 40 (height) square. The arrow should be
     * pointing down.
     * 2) UNSEEN_ELEMENT - will create a square of size 40.
     * 3) SEEN_ELEMENT - will create a GeneralPath that describes a rounded square of size 40.
     *
     * Please note that the numbers above are pure estimates and the programmer can change the defaults to more
     * appropriate values.
     */
    public ShapeFactory() {
        // Create a GeneralPath for Relationship Arrow.
        GeneralPath relationshipArrow = new GeneralPath();
        relationshipArrow.moveTo(14, 0);
        relationshipArrow.lineTo(14, 30);
        relationshipArrow.lineTo(19, 30);
        relationshipArrow.lineTo(10, 40);
        relationshipArrow.lineTo(0, 30);
        relationshipArrow.lineTo(5, 30);
        relationshipArrow.lineTo(5, 0);
        relationshipArrow.closePath();
        shapes.put(new Integer(ShapeFactory.RELATIONSHIP_ARROW), relationshipArrow);

        // Create a Rectangle for Unseen Element.
        GeneralPath unseenElement = new GeneralPath();
        unseenElement.moveTo(0, 0);
        unseenElement.lineTo(0, 39);
        unseenElement.lineTo(39, 39);
        unseenElement.lineTo(39, 0);
        unseenElement.closePath();
        shapes.put(new Integer(ShapeFactory.UNSEEN_ELEMENT), unseenElement);

        // Create a GeneralPath for Seen Element.
        GeneralPath seenElement = new GeneralPath();
        seenElement.moveTo(34, 0);
        seenElement.quadTo(38, 1, 39, 5);
        seenElement.lineTo(39, 34);
        seenElement.quadTo(38, 38, 34, 39);
        seenElement.lineTo(5, 39);
        seenElement.quadTo(1, 38, 0, 34);
        seenElement.lineTo(0, 5);
        seenElement.quadTo(1, 1, 5, 0);
        seenElement.closePath();
        shapes.put(new Integer(ShapeFactory.SEEN_ELEMENT), seenElement);
    }

    /**
     * This method allows the application to replace or add a new shape into the shape factory. If the shape replaces
     * an existing shape, that existing shape is returned. If the shape is a new shape, then null is returned. Please
     * be sure to convert the int to an Integer prior to hitting the shapes map.
     *
     * @param type An integer representing the unique type
     * @param shape The shape to relate to the type
     * @return The shape that was replaced or null if it was a new shape
     * @throws NullPointerException if type or shape is null
     * @throws IllegalArgumentException if shape does not implement the Clonable interface
     * @throws IllegalArgumentException if type is less than or equal to 0
     */
    public GeneralPath setShape(int type, GeneralPath shape) {
        if (type <= 0) {
            throw new IllegalArgumentException("type should be positive.");
        }
        if (shape == null) {
            throw new NullPointerException("shape should not be null.");
        }
        if (!(shape instanceof Cloneable)) {
            throw new IllegalArgumentException("shape should implement Clonable.");
        }
        return (GeneralPath) this.shapes.put(new Integer(type), shape);
    }

    /**
     * This method will return a shape that was cloned from the shape that was identified by the type. This method
     * simply retrieves the shape for the type from the shapes variable. If the shape was not found, return null.
     * If the shape was found, clone it (the shape will implement cloneable as enforced by the setShape method)
     * and return it. Please be sure to convert the int to an Integer prior to hitting the shapes map.
     *
     * @param type The type that identifies the shape
     * @return A new Shape that was cloned from the Shape identified by the type or null if the shape was not found
     * @throws IllegalArgumentException if type is less than or equal to 0
     */
    public GeneralPath getShape(int type) {
        if (type <= 0) {
            throw new IllegalArgumentException("type should be positive.");
        }
        GeneralPath shape = (GeneralPath) this.shapes.get(new Integer(type));
        if (shape == null) {
            return null;
        }

        return (GeneralPath) shape.clone();
    }

    /**
     * This method will remove the specified shape from the shape factory. The specified type cannot be on of the
     * common types (as defined by the constants). The shape that was removed is returned (or null if the type was
     * not found). Please be sure to convert the int to an Integer prior to hitting the shapes map.
     *
     * @param type The type that identifies the shape to remove
     * @return The  shape that was removed is returned (or null if the type was not found).
     * @throws IllegalArgumentException if type is less than or equal to 0
     * @throws IllegalArgumentException if the type is one of the predefined types
     */
    public GeneralPath removeShape(int type) {
        if (type <= 0) {
            throw new IllegalArgumentException("type should be positive.");
        }
        if (type == ShapeFactory.RELATIONSHIP_ARROW || type == ShapeFactory.UNSEEN_ELEMENT ||
            type == ShapeFactory.SEEN_ELEMENT) {
            throw new IllegalArgumentException("Pre-defined type should not be removed.");
        }
        return (GeneralPath) this.shapes.remove(new Integer(type));
    }

}
